produceOffspring		= MPI_NODE_COUNT-1;
if (produceOffspring <= 0)
{
	produceOffspring = 15;
}

populationSize  		= 2*produceOffspring;
incestDistance  		= 0;
generationCount		  	= 5000;
maxSampleTries			= populationSize*10;
mutationThreshhold		= 0.0001;
mutationProb			= 0.15;
mutationProbDecrease	= 0.95;
annealingPhase			= 100;
SHORT_MPI_RETURN		= 1;
totalSampleCounter		= 0;
localMutationRate		= 0.05;
localMutationInterval	= 20;

stoppingCriterion		= 100;
sampleCount				= 0;
familyControlSize		= produceOffspring$6;

verboseFlag				= 1;
stateVectorDimension    = 120;
rateClassesCount		= 4;

/* ________________________________________________________________________________________________*/

MasterList				= {};
REPLACE_TREE_STRUCTURE  = 1;
SHORT_MPI_RETURN		= 1;

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function StringToMatrix (zz)
{
	return zz;
}

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function getTipDatesFromNames (treeNameID)
{
	retAVL = {};
	ExecuteCommands ("ttipCount=TipCount("+treeNameID+");");
	regExpToMatch = "[0-9]*\\.?[0-9]+$";
	maxV = 0;
	for (tc = 0; tc < ttipCount; tc=tc+1)
	{
		ExecuteCommands ("abn = TipName("+treeNameID+",tc);");
		rgm = abn$regExpToMatch;
		if (rgm[0]<0)
		{
			return {};
		}
		si = rgm[0];
		ei = rgm[1];
		labelV = 0+abn[si][ei];
		if (labelV <= 0)
		{
			return {};
		}
		if (labelV > maxV)
		{
			maxV = labelV;
		}
		retAVL[abn] = labelV;
	}
	if (maxV>1)
	{
		maxV = 10^(-((Log(maxV)/Log(10)+0.5)$1));
		for (tc = 0; tc < ttipCount; tc=tc+1)
		{
			ExecuteCommands ("abn = TipName("+treeNameID+",tc);");
			retAVL[abn] = retAVL[abn]*maxV;
		}		
	}
	return retAVL;
}

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function ExportAMatrix (fileName, rateMatrix,dummy)
{
	if (dummy == 1)
	{
		return 0;
	}
	
	fprintf (fileName, CLEAR_FILE,ConvertToPartString(rateMatrix));
	return 0;
}

/*-----------------------------------------------------------*/

function generateDatedTipConstraints (treeNameID, parameterToConstrain, tipDateAVL, rateAVL)
{
	DT_String = "";
	DT_String * 8192;
	

	ExecuteCommands ("treePostOrderAVL = "+treeNameID+"^0;");
	nodeCount 	=  Abs(treePostOrderAVL);
	doneAssignment = {};
	
	for (nodeIndex = 0; nodeIndex < nodeCount-1; nodeIndex = nodeIndex+1)
	{
		aRate = rateAVL[nodeIndex];
		if (doneAssignment[aRate] == 0)
		{
			doneAssignment[aRate] = 1;
			DT_String * ("global "+treeNameID+"_scaler_"+aRate+" = 1.0;\n"+treeNameID+"_scaler_"+aRate+"_ :> 0.0;");
		}
	}

	for (nodeIndex = 1; nodeIndex < nodeCount; nodeIndex = nodeIndex+1)
	{
		nodeInfo 	= treePostOrderAVL[nodeIndex];
		nodeNameS	= nodeInfo["Name"];
	
		if (Abs(nodeInfo["Children"]))
		{
			DT_String * ("global "+treeNameID+"_"+nodeNameS+"_T = 1;\n");
		}
		else
		{
			nodeInfo["T"] =  tipDateAVL[nodeNameS];
			treePostOrderAVL[nodeIndex] = nodeInfo;
		}
	}
	
	for (nodeIndex = 1; nodeIndex < nodeCount; nodeIndex = nodeIndex+1)
	{
		nodeInfo 	= treePostOrderAVL[nodeIndex];
		nodeNameS	= nodeInfo["Name"];
		nodeParent = nodeInfo["Parent"];
		if (nodeParent)
		{
			parentNodeID = nodeParent;
			nodeParent = treePostOrderAVL[nodeParent];
		}
		
		if (Abs(nodeInfo["Children"]))
		{
			DT_String * (""+treeNameID+"_"+nodeNameS+"_T = "+nodeInfo["T"]+";\n");
		}
		
		if (Abs(nodeParent))
		{
			nodePT = nodeParent["T"];
			
			if (nodePT>0)
			{
				nodePT = Min(nodeInfo["T"]*Random(0.9,0.99), nodePT);
			}
			else
			{
				nodePT = nodeInfo["T"]*Random(0.9,0.99);
			}
			
			nodeParent["T"] = nodePT;
			treePostOrderAVL[parentNodeID] = nodeParent;
			
			
			rateClass = rateAVL[nodeIndex-1];
						
			if (Abs(nodeInfo["Children"]))
			{
				DT_String * (treeNameID+"."+nodeNameS+"."+parameterToConstrain+":="+treeNameID+"_scaler_"+rateClass+"*("+
							 treeNameID+"_"+nodeNameS+"_T-"+treeNameID+"_"+nodeParent["Name"]+"_T);\n");
			}
			else
			{
				DT_String * (treeNameID+"."+nodeNameS+"."+parameterToConstrain+":="+treeNameID+"_scaler_"+rateClass+"*("+
							 tipDateAVL[nodeNameS]+"-"+treeNameID+"_"+nodeParent["Name"]+"_T);\n");			
			}
			
		}
	}
	
	DT_String * 0;
	return DT_String;
}


/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function CleanUpMPI (dummy)
{
	if (MPI_NODE_COUNT>1)
	{
		while (1)
		{
			for (nodeCounter = 0; nodeCounter < MPI_NODE_COUNT-1; nodeCounter = nodeCounter+1)
			{
				if (MPINodeState[nodeCounter][0]==1)
				{
					fromNode = ReceiveJobs (0,0);
					break;	
				}
			}
			if (nodeCounter == MPI_NODE_COUNT-1)
			{
				break;
			}
		}			
	}
	return 0;
}

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function ReceiveJobs (sendOrNot, ji)
{
	if (MPI_NODE_COUNT>1)
	{
		MPIReceive (-1, fromNode, result_String);
		mji = MPINodeState[fromNode-1][1];
		
		if (sendOrNot)
		{
			MPISend (fromNode,lf);
			MPINodeState[fromNode-1][1] = ji;			
		}
		else
		{
			MPINodeState[fromNode-1][0] = 0;
			MPINodeState[fromNode-1][1] = -1;		
		}
		ExecuteCommands (result_String);
		myAIC = 2*(lf_MLES[1][0]-lf_MLES[1][1]-baseParams);
		ji = mji;
	}
	else
	{
		myAIC = 2*(res[1][0]-res[1][1]-baseParams);
	}
	
	if (resultProcessingContext==0)
	{
		sortedScores[ji][0] = myAIC;
		if (ji>=0)
		{
			jobPrint = ConvertToPartString (currentPopulation[ji]);
		}
		if (verboseFlag)
		{
			fprintf (stdout, "Individual ",ji," AIC = ",-myAIC," ");
		}
	}
	else
	{
		intermediateProbs[ji][0] = myAIC;	
		if (ji>=0)
		{
			jobPrint = ConvertToPartString (children[ji-populationSize]);
		}
		if (verboseFlag)
		{
			fprintf (stdout, "Offspring ",ji," AIC = ",-myAIC," ");
		}
	}
	
	MasterList [jobPrint] = myAIC;
	if (verboseFlag)
	{
		fprintf (stdout, " ", jobPrint, "\n");
	}
	
	return fromNode-1;
}

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function ConvertToPartString (modelSpec)
{
	outString = "";
	outString * 128;
	outString * (""+modelSpec[0]);
	
	for (zzz = 1; zzz < stateVectorDimension; zzz = zzz + 1)
	{
		outString * (","+modelSpec[zzz]);
	}
	
	outString * 0;
	return outString;
}

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function MakeStringCanonical (randomModel, classCount)
{
	compressedString = {classCount,1};
	for (h=0; h<stateVectorDimension; h=h+1)
	{
		v = randomModel[h];
		compressedString[v] = 1;
	}
	compressedString[0] = 0;
	for (h=1; h<classCount; h=h+1)
	{
		compressedString[h] = compressedString[h]+compressedString[h-1];
	}
	for (h=0; h<stateVectorDimension; h=h+1)
	{
		v = randomModel[h];
		v = compressedString[v];
		randomModel[h] = v;
	}
	v = compressedString[classCount-1]+1;
	if (v>1)
	{
		sortedOrder = {v,2};
		for (h=0; h<v; h=h+1)
		{
			sortedOrder[h][0] = -1;
		}
		cc = 0;
		for (h=0; h<stateVectorDimension; h=h+1)
		{
			hshift = randomModel[h];
			if (sortedOrder[hshift][0] < 0)
			{
				sortedOrder[hshift][0] = cc;
				sortedOrder[hshift][1] = hshift;
				cc = cc+1;
			}
		}
		sortedOrder = sortedOrder%1;
		for (h=0; h<stateVectorDimension; h=h+1)
		{
			v = randomModel[h];
			randomModel[h] = sortedOrder[v][0];
		}
		
	}
	return randomModel;
}


/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function RunASample (dummy,jobIndex)
{	
	DT_String = "";
	DT_String * 8192;
	

	ExecuteCommands ("treePostOrderAVL = "+treeNameID+"^0;");
	nodeCount 	=  Abs(treePostOrderAVL);

	myAIC = MasterList[cString];
	if (myAIC<0)
	{		
		if (resultProcessingContext==0)
		{
			sortedScores[jobIndex][0] = myAIC;
			if (verboseFlag)
			{
				fprintf (stdout, "Individual ",jobIndex," AIC = ",-myAIC, "\n");
			}
		}
		else
		{
			intermediateProbs[jobIndex][0] = myAIC;	
			if (verboseFlag)
			{
				fprintf (stdout, "Offspring ",jobIndex," AIC = ",-myAIC,"\n");
			}
		}	
	}
	else
	{
		ClearConstraints (clockTree);
		ExecuteCommands 				   (generateDatedTipConstraints  ("clockTree",parameter2ConstrainString,tipDateAVL,cString));
		LikelihoodFunction lf			 = (filteredData, clockTree);
		
		if ((MPI_NODE_COUNT>1) && (jobIndex>=0))
		{
			for (mpiNode = 0; mpiNode < MPI_NODE_COUNT-1; mpiNode = mpiNode+1)
			{
				if (MPINodeState[mpiNode][0]==0)
				{
					break;	
				}
			}
			if (mpiNode==MPI_NODE_COUNT-1)
			{
				mpiNode = ReceiveJobs (1,jobIndex);
			}
			else
			{
				MPISend (mpiNode+1,lf);
				MPINodeState[mpiNode][0] = 1;
				MPINodeState[mpiNode][1] = jobIndex;
			}
		}
		else
		{
			Optimize (res,lf);
			if (jobIndex>=0)
			{
				mpiNode = ReceiveJobs (1, jobIndex);
			}
			else
			{
				myAIC = 2*(res[1][0]-res[1][1]-baseParams);
			}
		}
	}
	return 0;	
}




/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function SpawnRandomString (clsCnt)
{
	rModel = {stateVectorDimension,1};
	for (h=0; h<stateVectorDimension; h=h+1)
	{
		rModel[h] = Random(0,rateClassesCount)$1;
	}
	return MakeStringCanonical(rModel,rateClassesCount);
}

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function IsChildViable (putativeChild)
{
	sampleString = 	putativeChild;

	myAIC = MasterList[sampleString];
	testChild = putativeChild;
	mutPassCount = 1;
	while (myAIC < (-0.1) && mutPassCount < 25)
	{
		if (verboseFlag)
		{
			fprintf (stdout,"Adjusting the child to avoid a duplicate. Pass ", mutPassCount, "\n");
		}
		
		mutPassCount = mutPassCount + 1;
		
		sampleString = Min(Random(0,stateVectorDimension)$1,stateVectorDimension-1);
		myAIC = testChild[sampleString];
		
		newValue = Random (0,rateClassesCount-0.0000001)$1;
		
		while (newValue == myAIC)
		{
			newValue = Random (0,rateClassesCount-0.0000001)$1;
		}
		
		testChild [sampleString] = newValue;
		myAIC = MasterList[testChild];
	}
	return testChild;
}

/*---------------------------------------------------------------------------------------------------------------------------------------------*/

function UpdateBL (dummy)
{
	return 0;
}


/*---------------------------------------------------------------------------------------------------------------------------------------------*/


ChoiceList (dataType,"Data type",1,SKIP_NONE,"Nucleotide","Nucleotide data.",
				     "Codon","Codon (several available genetic codes).");

if (dataType<0) 
{
	return;
}

if (dataType==0)
{
	SetDialogPrompt 	  		  ("Locate a nucleotide data  file:");
}
else
{
	SetDialogPrompt 	  		  ("Locate a codon data file:");
}

DataSet 		ds  		 = ReadDataFile (PROMPT_FOR_FILE);
DataSetFilter	filteredData = CreateFilter(ds,1);

fprintf (stdout, "How many rates?");
fscanf (stdin, "Number", rateClassesCount);

SelectTemplateModel (filteredData);

incFileName = HYPHY_BASE_DIRECTORY+"TemplateBatchFiles"+DIRECTORY_SEPARATOR+"queryTree.bf";
ExecuteCommands  ("#include \""+incFileName+"\";");

tipDateAVL = getTipDatesFromNames ("givenTree");
if (Abs(tipDateAVL) == 0)
{
	fprintf (stdout, "\nERROR: All sequence names should end with numeric dates, e.g. env2898sampled96\n");
	return 0;
}

parameter2Constrain = 0;

ACCEPT_ROOTED_TREES = 1;
Tree clockTree = treeString;

if (Rows("LAST_MODEL_PARAMETER_LIST")>1)
{
	ChoiceList (parameter2Constrain, "Parameter(s) to constrain:",1,SKIP_NONE,LAST_MODEL_PARAMETER_LIST);

	if (parameter2Constrain<0)
	{
		return;
	}
	if (parameter2Constrain==0)
	{
		parameter2ConstrainString = "";
		for (parameter2Constrain=Rows("LAST_MODEL_PARAMETER_LIST")-1; parameter2Constrain; parameter2Constrain = parameter2Constrain-1)
		{
			GetString (funnyString,LAST_MODEL_PARAMETER_LIST,parameter2Constrain);
			parameter2ConstrainString = parameter2ConstrainString + funnyString + ",";
		}
		GetString (funnyString,LAST_MODEL_PARAMETER_LIST,0);
		parameter2ConstrainString = parameter2ConstrainString + funnyString;
	}
	else
	{
		GetString (parameter2ConstrainString,LAST_MODEL_PARAMETER_LIST,parameter2Constrain-1);
	}
}
else
{
	GetString (parameter2ConstrainString,LAST_MODEL_PARAMETER_LIST,0);
}

SetDialogPrompt ("Save the best model to:");
fprintf (PROMPT_FOR_FILE,CLEAR_FILE);
modelFile = LAST_FILE_PATH;

Tree givenTree = treeString;
stateVectorDimension = Columns(BranchLength (clockTree,-1));

LikelihoodFunction lf = (filteredData,givenTree);
Optimize (res,lf);

baseParams = res[1][2];
	
if (baseParams>0)
{
	ConstraintString = "";
	ConstraintString*256;
	for (h=0; h<baseParams; h=h+1)
	{
		GetString (v,lf,h);
		ConstraintString * (v+":="+v+"__;\n");
	}
	ConstraintString*0;
	ExecuteCommands (ConstraintString);
}


Tree clockTree = treeString;
cString = {stateVectorDimension,1};
ExecuteCommands 	(generateDatedTipConstraints  ("clockTree",parameter2ConstrainString,tipDateAVL,cString));
LikelihoodFunction lf = (filteredData,clockTree);
Optimize (res,lf);

currentPopulation  = {};
sortedScores	   = {populationSize,2};
sortedScores[0][0] = 2*(res[1][0]-res[1][1]);
sortedScores[0][1] = 0;

currentPopulation [0] = cString;
fprintf (stdout, "\n______________SINGLE RATE______________\n",lf,"\nAIC=",-sortedScores[0][0],"\n");


ibfPath = HYPHY_BASE_DIRECTORY + "TemplateBatchFiles" + DIRECTORY_SEPARATOR + "GA_CHC.ibf";
ExecuteCommands ("#include \"" + ibfPath + "\";");


ExportAMatrix (modelFile,currentPopulation[populationSize-1],0);
modelFile = modelFile + ".samples";

masterKeys = Rows(MasterList);
outString = "";
outString * 65536;
for (h=Rows(masterKeys)*Columns(masterKeys)-1; h>=0; h=h-1)
{
	aKey = masterKeys[h];
	outString * (""+(-MasterList[aKey])+"\n");
	outString * aKey;
	outString * "\n";
}
outString * 0;
fprintf (modelFile,CLEAR_FILE,outString);

